-- Big value keeper
RAREAN = {}
-- Main refreshing and utility frame
local mainFrame  = CreateFrame("Frame", nil, UIParent)

-- ToDo:
-- Interface options menu with options for: (Interface option menu done, need to add the rest of the options)
--   1) Respond to raretime / raretimer requests (Implemented)
--   2) Report rare finds when targeting a rare or when its killed closeby (Implemented)
--   3) Hide window when in combat
--   4) Choose method of rare tracking:
--      a) All
--      b) Selected
--      c) Next to spawn (with max, where 0 == all that are ready)
--   4.1) Show all rares on mouse-over
--   5) Set TomTom waypoint on rare detection from other people (only for the rares selected in option 3)
--   6) Order rare list by:
--      a) Name
--      b) Last kill time
-- TomTom waypoints when a rare is found, but you are not targetting it
-- 


-- The global options stored cross sessions
RAREAN.Options = {
  RespondToRequests,
  ReportFindsToGeneral,
  HideInCombat,
  ShowAllOnMouseOver,
  SetTomTomWaypoints,
  RareOrderType,
  TrackingType,
  TrackingRares = {}
}

-- The default options
local OptionsDefault = {
  RespondToRequests = true,
  ReportFindsToGeneral = true,
  HideInCombat = true,
  ShowAllOnMouseOver = true,
  SetTomTomWaypoints = true,
  RareOrderType = 1, -- 1 = Name, 2 = Time
  TrackingType = 1, -- 1 = All, 2 = Selected, 3 = Next to spawn
  TrackingRares = {}
}

local mobIDs = {
  73174, -- Archiereus of Flame
  72775, -- Bufo
  73171, -- Champion of the Black Flame
  72045, -- Chelon
  73175, -- Cinderfall
  72049, -- Cranegnasher
  73281, -- Dread Ship Vazuvius
  73158, -- Emerald Gander
  73279, -- Evermaw
  73172, -- Flintlord Gairan
  73282, -- Garnia
  72970, -- Golganarr
  73161, -- Great Turtle Furyshell
  72909, -- Gu'chi the Swarmbringer
  73167, -- Huolon
  73163, -- Imperial Python
  73160, -- Ironfur Steelhorn
  73169, -- Jakur of Ordon
  72193, -- Karkanos
  73277, -- Leafmender
  73166, -- Monsterous Spineclaw
  72048, -- Rattleskew
  73157, -- Rock Moss
  71864, -- Spelurk
  72769, -- Spirit of Jadefire
  73704, -- Stinkbraid
  72808, -- Tsavo'ka <Ghost in the Darkness>
  73173, -- Urdur the Cauterizer
  73170, -- Watcher Osu
  72245, -- Zesqua
  71919, -- Zhu-Gon the Sour
  --- Testing
  --73162, -- Foreboding Flame
}

local mobs = {
  -- npc_id = {message sent, died time, name, min spawn time, max spawn time, coordsX, coordsY, lastRequest}
  -- Timeless Isle (map 951)
  -- Many of the spawn timers are guesses, will need to refine when data is available
  [73174] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Archiereus of Flame"         , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Archiereus of Flame
  [72775] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Bufo"                        , minSpawn = 1800, maxSpawn = 3600, coordsX = 64.6, coordsY = 72.8, lastRequest = 0 }, -- Bufo
  [73171] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Champion of the Black Flame" , minSpawn = 1800, maxSpawn = 3600, coordsX = 61.6, coordsY = 48.4, lastRequest = 0 }, -- Champion of the Black Flame
  [72045] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Chelon"                      , minSpawn = 1800, maxSpawn = 3600, coordsX = 25.2, coordsY = 36.0, lastRequest = 0 }, -- Chelon
  [73175] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Cinderfall"                  , minSpawn = 1800, maxSpawn = 3600, coordsX = 54.6, coordsY = 53,   lastRequest = 0 }, -- Cinderfall
  [72049] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Cranegnasher"                , minSpawn = 1800, maxSpawn = 3600, coordsX = 43.8, coordsY = 69.6, lastRequest = 0 }, -- Cranegnasher
  [73281] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Dread Ship Vazuvius"         , minSpawn = 1800, maxSpawn = 3600, coordsX = 25.8, coordsY = 23.2, lastRequest = 0 }, -- Dread Ship Vazuvius
  [73158] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Emerald Gander"              , minSpawn =  900, maxSpawn = 1800, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Emerald Gander
  [73279] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Evermaw"                     , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Evermaw
  [73172] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Flintlord Gairan"            , minSpawn = 1800, maxSpawn = 3600, coordsX = 47.8, coordsY = 38.8, lastRequest = 0 }, -- Flintlord Gairan
  [73282] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Garnia"                      , minSpawn = 1800, maxSpawn = 3600, coordsX = 63.4, coordsY = 28.4, lastRequest = 0 }, -- Garnia
  [72970] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Golganarr"                   , minSpawn = 1800, maxSpawn = 3600, coordsX = 61.8, coordsY = 63,   lastRequest = 0 }, -- Golganarr
  [73161] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Great Turtle Furyshell"      , minSpawn =  900, maxSpawn = 1800, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Great Turtle Furyshell
  [72909] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Gu'chi the Swarmbringer"     , minSpawn = 1800, maxSpawn = 3600, coordsX = 40.6, coordsY = 81.6, lastRequest = 0 }, -- Gu'chi the Swarmbringer
  [73167] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Huolon"                      , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Huolon
  [73163] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Imperial Python"             , minSpawn =  900, maxSpawn = 1800, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Imperial Python
  [73160] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Ironfur Steelhorn"           , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Ironfur Steelhorn
  [73169] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Jakur of Ordon"              , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Jakur of Ordon
  [72193] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Karkanos"                    , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Karkanos
  [73277] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Leafmender"                  , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Leafmender
  [73166] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Monstrous Spineclaw"         , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Monsterous Spineclaw
  [72048] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Rattleskew"                  , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Rattleskew
  [73157] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Rock Moss"                   , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Rock Moss
  [71864] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Spelurk"                     , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Spelurk
  [72769] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Spirit of Jadefire"          , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Spirit of Jadefire
  [73704] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Stinkbraid"                  , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Stinkbraid
  [72808] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Tsavo'ka"                    , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Tsavo'ka <Ghost in the Darkness>
  [73173] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Urdur the Cauterizer"        , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Urdur the Cauterizer
  [73170] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Watcher Osu"                 , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Watcher Osu
  [72245] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Zesqua"                      , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Zesqua
  [71919] = { lastAnnounce = 0, killTime = -1, killAnnounced = 0, npcName = "Zhu-Gon the Sour"            , minSpawn = 1800, maxSpawn = 3600, coordsX = 0,    coordsY = 0,    lastRequest = 0 }, -- Zhu-Gon the Sour
  -- Testing
  --[73162] = { lastAnnounce = 0, killTime = -1, npcName = "Foreboding Flame"            , minSpawn = 2,    maxSpawn = 5,    coordsX = 0, coordsY = 0, lastRequest = 0 }, -- Foreboding Flame
}

-- local text_timer, text_throttle = 0, 10
local textRefreshThrottle = 5 -- Refreshing every 5 secs
local textRefreshCurrent  = 0
local channelJoinDelay = 5 -- Delay joining and syncing 5s after load, prevents our channel to become one of the first
local channelJoinCurrent = 0
local needSync = false
local needRCSync = false
local syncChanNamePrefix = "RAREAN"
local syncRSChanNamePrefix = "RCELVA"

-- Main Frame Design:
-- |-------- MainFrame -------------------------------------------|
-- ||------- NPC Name -------------------------|----  Killed ----||
-- || Some NPC Name                            |       1min      ||
-- || Some NPC Name                            |      15min      ||
-- || Some NPC Name                            |       1min      ||
-- || Some NPC Name                            |       1min      ||
-- || Some NPC Name                            |       1min      ||
-- || Some NPC Name                            |       1min      ||
-- || Some NPC Name                            |       1min      ||
-- ||------------------------------------------------------------||
-- |--------------------------------------------------------------|

local rareFrame = CreateFrame("Frame", "rareFrame", UIParent)
  rareFrame:SetWidth(261)
  rareFrame:SetHeight(34)
  rareFrame:SetPoint("LEFT", 300, 100)
  rareFrame:SetMovable(true)
  rareFrame:EnableMouse(true)
  rareFrame:RegisterForDrag("LeftButton")
  rareFrame:SetScript("OnDragStart", mainFrame.StartMoving)
  rareFrame:SetScript("OnDragStop", mainFrame.StopMovingOrSizing)
  rareFrame:SetUserPlaced(true)
  rareFrame.texture = rareFrame:CreateTexture(nil, "BACKGROUND")
  rareFrame.texture:SetTexture(0,0,0,0.4)
  rareFrame.texture:SetAllPoints(rareFrame)

  rareFrame.npcFrame = CreateFrame("Frame", "rareFrame.npcFrame", rareFrame)
  rareFrame.npcFrame:SetWidth(180)
  rareFrame.npcFrame:SetHeight(rareFrame:GetHeight()-10)
  rareFrame.npcFrame:SetPoint("TOPLEFT", rareFrame, 5, -5)
  rareFrame.npcFrame.texture = rareFrame.npcFrame:CreateTexture(nil, "BACKGROUND")
  rareFrame.npcFrame.texture:SetTexture(0,0,0,0.5)
  rareFrame.npcFrame.texture:SetAllPoints(rareFrame.npcFrame)

  rareFrame.timeFrame = CreateFrame("Frame", "rareFrame.timeFrame", rareFrame)
  rareFrame.timeFrame:SetWidth(70)
  rareFrame.timeFrame:SetHeight(rareFrame:GetHeight()-10)
  rareFrame.timeFrame:SetPoint("TOPLEFT", rareFrame.npcFrame, "TOPRIGHT", 1, 0)
  rareFrame.timeFrame.texture = rareFrame.timeFrame:CreateTexture(nil, "BACKGROUND")
  rareFrame.timeFrame.texture:SetTexture(0,0,0,0.5)
  rareFrame.timeFrame.texture:SetAllPoints(rareFrame.timeFrame)

rareFrame.npcFrame.text = {}
local i
local validCount = 0
for i=0, table.getn(mobIDs) do
  if (i == 0) then
    -- Header
    rareFrame.npcFrame.text[i] = rareFrame.npcFrame:CreateFontString("rareFrame.npcFrame.text["..i.."]", nil, "GameFontNormal")
    rareFrame.npcFrame.text[i]:SetPoint("TOP", "rareFrame.npcFrame", 2, -2 + -14.*i)
    rareFrame.npcFrame.text[i]:SetFont("Fonts\\ARIALN.TTF", 12)
    rareFrame.npcFrame.text[i]:SetTextColor(1, 0.7, 0, 1)
    rareFrame.npcFrame.text[i]:SetText("Name")
    rareFrame.npcFrame.text[i]:SetFont("Fonts\\ARIALN.TTF", 12,"OUTLINE")
  elseif (i > 0) then
    if mobIDs[i] ~= nil then
      if (mobs[mobIDs[i]].npcName ~= nil) then
        validCount = validCount + 1
        rareFrame.npcFrame.text[validCount] = rareFrame.npcFrame:CreateFontString("rareFrame.npcFrame.text["..validCount.."]", nil, "GameFontNormal")
        rareFrame.npcFrame.text[validCount]:SetPoint("TOP", "rareFrame.npcFrame", 2, 0 + -14.* validCount)
        rareFrame.npcFrame.text[validCount]:SetFont("Fonts\\ARIALN.TTF", 12)
        rareFrame.npcFrame.text[validCount]:SetTextColor(1, 0.8, 0, 1)
        rareFrame.npcFrame.text[validCount]:SetText(mobs[mobIDs[i]].npcName)
        rareFrame.npcFrame.text[validCount]:SetJustifyH("LEFT")
      end
    end
  end
end

rareFrame.timeFrame.text = {}
local i
local validCount = 0
for i=0, table.getn(mobIDs) do
  if (i == 0) then
    -- Header
    rareFrame.timeFrame.text[i] = rareFrame.timeFrame:CreateFontString("rareFrame.timeFrame.text["..i.."]", nil, "GameFontNormal")
    rareFrame.timeFrame.text[i]:SetPoint("TOP", "rareFrame.timeFrame", 2, -2 + -14.*i)
    rareFrame.timeFrame.text[i]:SetFont("Fonts\\ARIALN.TTF", 12)
    rareFrame.timeFrame.text[i]:SetTextColor(1, 0.7, 0, 1)
    rareFrame.timeFrame.text[i]:SetText("Killed")
    rareFrame.timeFrame.text[i]:SetFont("Fonts\\ARIALN.TTF", 12,"OUTLINE")
  elseif (i > 0) then
    if mobIDs[i] ~= nil then
      if (mobs[mobIDs[i]].npcName ~= nil) then
        validCount = validCount + 1
        rareFrame.timeFrame.text[validCount] = rareFrame.timeFrame:CreateFontString("rareFrame.timeFrame.text["..validCount.."]", nil, "GameFontNormal")
        rareFrame.timeFrame.text[validCount]:SetPoint("TOP", "rareFrame.timeFrame", 3, 0 + -14.* validCount)
        rareFrame.timeFrame.text[validCount]:SetFont("Fonts\\ARIALN.TTF", 12)
        rareFrame.timeFrame.text[validCount]:SetTextColor(1, 0.8, 0, 1)
        rareFrame.timeFrame.text[validCount]:SetText("---")
      end
    end
  end
end

local syncRareIte
local responseFormater = CreateFrame("Button")
responseFormater:Hide()

-- Converts the mobs index into a string value
local function GetRareIdString(rareID)
  local i
  for i = 1, table.getn(mobIDs) do
    if (mobIDs[i] == rareID) then
      return mobIDs[i]
    end
  end
end

-- Gets the channel id of a named custom channel
local function GetChannelID(channelName)
  local channelCount = GetNumDisplayChannels()
  
  for i = 1, channelCount do
    SetSelectedDisplayChannel(i)
    local ciName, _, _, _, _, _, _, _, _ = GetChannelDisplayInfo(i)
    if (ciName == channelName) then
      return i
    end
  end
end

-- Gets the owner of a given channel
local function GetChannelOwner(chanID)
  local _, _, _, _, memberCount, _, _, _, _ = GetChannelDisplayInfo(chanID)
  
  for i = 1, memberCount do
    local name, owner, _, _, _, _ = GetChannelRosterInfo(chanID, i)
    if (owner) then
      return name
    end
  end
end

-- Joins the different syncing channels (Currently ours and RareCoordinators)
local function JoinSyncChannels(self, elapsed)
  channelJoinCurrent = channelJoinCurrent + elapsed
  if (channelJoinCurrent > channelJoinDelay) then
    -- RareAnnouncer sync channel
    if (IsAddonMessagePrefixRegistered(syncChanNamePrefix) == false) then
      RegisterAddonMessagePrefix(syncChanNamePrefix)
    end
    
    -- Now find our syncing channel
    local syncChanID = GetChannelID(syncChanNamePrefix)
    if (syncChanID == nil) then
      JoinChannelByName(syncChanNamePrefix)
    end
    needSync = true
    
    -- RareCoordinator sync channel
    if (IsAddonMessagePrefixRegistered(syncRSChanNamePrefix) == false) then
      RegisterAddonMessagePrefix(syncRSChanNamePrefix)
    end
    
    -- Now find RC syncing channel
    local syncRCChanID = GetChannelID(syncRSChanNamePrefix)
    if (syncRCChanID == nil) then
      JoinChannelByName(syncRSChanNamePrefix)
    end
    needRCSync = true
    
    channelJoinCurrent = 0
  end
end

-- Refreshes the main text window
local function RefreshText()
  if (rareFrame.npcFrame ~= nil) then
    if (rareFrame.npcFrame.text ~= nil) then
      -- Loop through our rare mobs, check if we have a name and kill time, then search for it in npcFrame
      -- if it is there, update the time, otherwise add it
      local textIndex = 0
      local rareIndex = 0
      local textFound = false
      for rareIndex = 0, table.getn(mobIDs) do
        if (mobs[mobIDs[rareIndex]] ~= nil) then
          if (mobs[mobIDs[rareIndex]].npcName ~= nil) then
            textFound = false

            for textIndex = 0, table.getn(rareFrame.npcFrame.text) do
              if (rareFrame.npcFrame.text[textIndex]:GetText() == mobs[mobIDs[rareIndex]].npcName) then
                textFound = true
                if (mobs[mobIDs[rareIndex]].killTime ~= nil) then
                  local timeSinceKill = (time() - mobs[mobIDs[rareIndex]].killTime)
                  local mediumSpawn = (mobs[mobIDs[rareIndex]].minSpawn + mobs[mobIDs[rareIndex]].maxSpawn) / 2
                  
                  if (mobs[mobIDs[rareIndex]].killTime == -1) then
                    rareFrame.timeFrame.text[textIndex]:SetText("---")
                    rareFrame.timeFrame.text[textIndex]:SetTextColor(1, 0.8, 0, 1) -- Normal
                  elseif (timeSinceKill < 60) then
                    rareFrame.timeFrame.text[textIndex]:SetText("Dead")
                    rareFrame.timeFrame.text[textIndex]:SetTextColor(1, 0.8, 0, 1) -- Normal
                  else
                    rareFrame.timeFrame.text[textIndex]:SetText(SecondsToTime(timeSinceKill, true))
                    
                    -- Set the text colour
                    -- KillTime < Minimum Spawn time
                    if (timeSinceKill < mobs[mobIDs[rareIndex]].minSpawn) then
                      rareFrame.timeFrame.text[textIndex]:SetTextColor(1, 0.8, 0, 1) -- Normal
                    -- KillTime > Minimum Spawn time, but < Maximum Spawn time
                    -- and < 1/2 of the Average between Min and Max
                    elseif ((timeSinceKill > mobs[mobIDs[rareIndex]].minSpawn) and
                             (timeSinceKill < mediumSpawn)) then
                      rareFrame.timeFrame.text[textIndex]:SetTextColor(0, 1, 0, 1)  -- Green
                    -- KillTime > Minimum Spawn time, but < Maximum Spawn time
                    -- and > 1/2 of the Average between Min and Max
                    elseif ((timeSinceKill >= mediumSpawn) and
                            (timeSinceKill < mobs[mobIDs[rareIndex]].maxSpawn)) then
                      rareFrame.timeFrame.text[textIndex]:SetTextColor(0, 0.7, 0, 1) -- Light Green
                    -- KillTime > Maximum Spawn time. Either kill missed or no announcement
                    else
                      rareFrame.timeFrame.text[textIndex]:SetTextColor(1, 0, 0, 0.9) -- Red
                    end
                  end
                else
                  rareFrame.timeFrame.text[textIndex]:SetText("Alive")
                  rareFrame.timeFrame.text[textIndex]:SetTextColor(0, 0.8, 1, 1) -- Green
                end
                rareFrame.timeFrame.text[textIndex]:SetJustifyH("LEFT")
              end
            end

            -- First time the rare has been killed, so add it to the frame
            if (textFound == false) then
              local newIndex = table.getn(rareFrame.npcFrame.text) + 1
              rareFrame.npcFrame.text[newIndex] = rareFrame.npcFrame:CreateFontString("rareFrame.npcFrame.text["..newIndex.."]", nil, "GameFontNormal")
              rareFrame.npcFrame.text[newIndex]:SetPoint("TOP", "rareFrame.npcFrame", 2, -0 + -14.* newIndex)
              rareFrame.npcFrame.text[newIndex]:SetFont("Fonts\\ARIALN.TTF", 12)
              rareFrame.npcFrame.text[newIndex]:SetTextColor(1, 0.8, 0, 1)
              rareFrame.npcFrame.text[newIndex]:SetText(mobs[mobIDs[rareIndex]].npcName)
              rareFrame.npcFrame.text[newIndex]:SetJustifyH("LEFT")

              rareFrame.timeFrame.text[newIndex] = rareFrame.timeFrame:CreateFontString("rareFrame.timeFrame.text["..newIndex.."]", nil, "GameFontNormal")
              rareFrame.timeFrame.text[newIndex]:SetPoint("TOP", "rareFrame.timeFrame", 3, 0 + -14.* newIndex)
              rareFrame.timeFrame.text[newIndex]:SetFont("Fonts\\ARIALN.TTF", 12)
              rareFrame.timeFrame.text[newIndex]:SetTextColor(1, 0.8, 0, 1)
              if (mobs[mobIDs[rareIndex]].killTime ~= nil) then
                if (mobs[mobIDs[rareIndex]].killTime == -1) then
                  rareFrame.timeFrame.text[newIndex]:SetText("---")
                else
                  rareFrame.timeFrame.text[newIndex]:SetText(SecondsToTime(time() - mobs[mobIDs[rareIndex]].killTime, true))
                end
              else
                rareFrame.timeFrame.text[newIndex]:SetText("Killed")
              end
            end
          end
        end
      end
      rareFrame.npcFrame:SetHeight(25 +  table.getn(rareFrame.npcFrame.text) *14)
      rareFrame.timeFrame:SetHeight(25 +  table.getn(rareFrame.npcFrame.text) *14)
      rareFrame:SetHeight(35 +  table.getn(rareFrame.npcFrame.text) *14)
    end
  end
end

-- General OnEvent handler (Refresher)
local function OverallUpdater(self,elapsed)
  textRefreshCurrent = textRefreshCurrent + elapsed
  if textRefreshCurrent >= textRefreshThrottle then
    -- Refresh the text frames
    RefreshText()
    textRefreshCurrent = 0
  end
end

-- Sends a message to either /1 or our syncing channel
local function SendGeneralMessage(msg, isSync)
  if (msg ~= nil) then
    if (isSync == false) then
      -- For some reason the SendChatMessage does not like the output from SecondsToTime
      -- this is a work around
      msg = responseFormater:GetText(responseFormater:SetText(msg))
      SendChatMessage(msg , "CHANNEL", nil, 1)
      -- SendChatMessage(msg , "CHANNEL", nil, 4) -- Testing
    else
      local syncChanID = GetChannelID(syncChanNamePrefix)
      if (syncChanID ~= nil) then
        SendChatMessage(msg, "CHANNEL", nil, syncChanID)
      end
    end
  end
end

-- Creates and sends the death message and sync
local function SendDeathMessage(rareID)
  if (mobs[rareID] ~= nil) then
    -- First the chat deathMsg
    local deathMsg
    if (RAREAN.Options.ReportFindsToGeneral) then
      deathMsg = "RareAnnouncer: ".. mobs[rareID].npcName .." has died"
      -- Kill announcement thottle
      if ((time() - mobs[rareID].killAnnounced) > 120) then
        SendGeneralMessage(deathMsg, false)
      end
    end
    
    -- Now the sync message
    deathMsg = "[".. syncChanNamePrefix .. "]" .. GetRareIdString(rareID) .. "_kill_" .. tostring(mobs[rareID].killTime)
    SendGeneralMessage(deathMsg, true)
    deathMsg = "[".. syncChanNamePrefix .. "]" .. GetRareIdString(rareID) .. "_killannounced"
    SendGeneralMessage(deathMsg, true)
  end
end

-- Creates and sends the target messages and sync
local function SendTargetMessage(rareID)
  if (RAREAN.Options.ReportFindsToGeneral == true) then
    if (mobs[rareID] ~= nil) then
      -- Only announce once every 10 seconds
      if (mobs[rareID].lastAnnounce + 10 < time()) then
        local name = strtrim(UnitName("target"))
        local x, y = GetPlayerMapPosition("player")
        local health = math.floor(UnitHealth("target")*100/UnitHealthMax("target"))
        
        local targetMessage = "RareAnnouncer: ".. mobs[rareID].npcName .." (".. health .."%). Coordinates: "..math.floor(x*100)..", "..math.floor(y*100)
        SendGeneralMessage(targetMessage, false)
        mobs[rareID].lastAnnounce = time()
        targetMessage = "[".. syncChanNamePrefix .. "]" ..GetRareIdString(rareID) .. "_target_" .. mobs[rareID].lastAnnounce
      end
    end
  end
end

-- Creates and sends the request response message and sync
local function SendRequestMessage(rareID)
  if (RAREAN.Options.RespondToRequests == true) then
    if (mobs[rareID] ~= nil) then
      -- We throttle the number of requests per mob to once every 1 minute
      if ((time() - mobs[rareID].lastRequest) > 60) and
          (mobs[rareID].npcName ~= nil) and
          (mobs[rareID].killTime ~= nil) and
          (mobs[rareID].killTime ~= -1) then
        
        local responseString
        
        -- Response in General
        if ((time() - mobs[rareID].killTime) > 60) then
          responseString = "RareAnnouncer: " .. mobs[rareID].npcName .. " was last killed " .. tostring(SecondsToTime(time() - mobs[rareID].killTime, true)) .. " ago."
        else
          responseString = "RareAnnouncer: " .. mobs[rareID].npcName .. " was last killed less than a minute ago."
        end
        SendGeneralMessage(responseString, false)
  
        -- Sync message
        mobs[rareID].lastRequest = time()
        responseString = "[".. syncChanNamePrefix .. "]" ..GetRareIdString(rareID) .. "_request_" .. mobs[rareID].lastRequest
        SendGeneralMessage(responseString, true)
      end
    end
  end
end

-- Request handler
local function ProcessRequest(message)
  if (RAREAN.Options.RespondToRequests == true) then
    -- Expected format = raretime <npcname / id>
    if (message ~= nil) then
      local searchingString
      
      if (strlower(string.sub(message,1,9)) == "raretime ") then
        searchingString = string.sub(message, 10, string.len(message))
      elseif (strlower(string.sub(message,1,10)) == "raretimer ") then
        searchingString = string.sub(message, 11, string.len(message))
      else
        return
      end
    
      if (string.len(searchingString) > 3) then
        local responseString
    
        -- First check for ID
        local searchID = tonumber(searchingString)
        if (searchID ~= nil) then
          if (mobs[searchID] ~= nil) then
            SendRequestMessage(searchID)
          end
        -- Now search on name or part of name
        else
          searchingString = strlower(searchingString)
          local i
          for i= 1, table.getn(mobIDs) do
            if (mobs[mobIDs[i]] ~= nil) then
              if (strlower(mobs[mobIDs[i]].npcName) == searchingString) or
                 (string.find(strlower(mobs[mobIDs[i]].npcName), searchingString) ~= nil) then
                SendRequestMessage(mobIDs[i])
                break
              end
            end
          end
        end
      end
    end
  end
end

-- Deathlog handler
local function ProcessDeathLog(name, guid)
  name = strtrim(name)
  local id = tonumber(guid:sub(6, 10), 16)
  
  if (mobs[id] ~= nil) then
    mobs[id].killTime = time()
    mobs[id].npcName = name
    SendDeathMessage(id)
  end
end

-- Our Sync channel message handler
local function ProcessSyncChatMsg(message)
  local prefix = "[" .. syncChanNamePrefix .. "]"
  local dataStr = string.sub(message, string.len(prefix)+1, string.len(message))
  
  local rareIDStr, syncType, syncTimeStr = strsplit("_", dataStr)
  local rareID = tonumber(strtrim(rareIDStr))
  local syncTime = tonumber(strtrim(syncTimeStr))
  
  if ((rareID       ~= nil) and
      (syncType     ~= nil) and
      (syncTime ~= nil)) then
    if (mobs[rareID] ~= nil) then

      if (syncType == "request") then
        mobs[rareID].lastRequest = syncTime
      elseif (syncType == "kill") or
             (syncType == "death") then
        mobs[rareID].killTime = syncTime
      elseif (syncType == "target") then
        mobs[rareID].lastAnnounce = syncTime
      elseif (syncType == "killannounced") then
        mobs[rareID].killAnnounced = syncTime
      end
    end
  end
end

-- RareCoordinator sync channel message handler
local function ProcessRSSyncChatMsg(message)
  local prefix = "[" .. syncChanNamePrefix .. "]"
  local dataStr
  if string.find(message, prefix) then
    dataStr = string.sub(message, string.len(prefix)+1, string.len(message))
  else
    dataStr = message
  end
  
  local rcVersion, rcRareIDStr, rcType, rcTimeStr = strsplit("_", dataStr)
  local rcRareID = tonumber(rcRareIDStr)
  local rcTime   = tonumber(rcTimeStr)
  -- print ("rcVersion=", rsVersion, " rcRareID=", rcRareID, " rcType=", rcType, " rcTimeStr=", rcTimeStr)
  
  if ((rcVersion ~= nil) and
      (rcRareID  ~= nil) and
      (rcType    ~= nil) and
      (rcTime    ~= nil)) then
    if (rcType == "killed") then
      if (mobs[rcRareID] ~= nil) then
        -- Only update if the kill time is less than ours
        if (mobs[rcRareID].killTime ~= nil) then
          -- print ("Rare " , mobs[rcRareID].npcName, " killTime is ", mobs[rcRareID].killTime, " rcTime is ", rcTime)
          if (mobs[rcRareID].killTime > rcTime) and
            (rcTime > 0) then
            mobs[rcRareID].killTime = rcTime
          elseif (mobs[rcRareID].killTime == -1) then
            mobs[rcRareID].killTime = rcTime
          end
        else
          -- print ("Rare " , mobs[rcRareID].npcName, " killTime is nil")
          mobs[rcRareID].killTime = rcTime
        end
      end
    elseif (rcType == "alive") then
      if (mobs[rcRareID] ~= nil) then
        mobs[rcRareID].killTime = nil
      end
    end
  end
end

-- The hidden addon channel message handler
local function ProcessAddonMsg(prefix, message, channel, sender)
  --print ("prefix = ", prefix, " channel = ", channel, " sender = ", sender)
  --print ("message = ", message)
  if (prefix == syncChanNamePrefix)then
    if (channel == "WHISPER") then
      -- Sync request
      if (message == "Sync") then
        for syncRareIte = 1, table.getn(mobIDs) do
          if (mobs[mobIDs[syncRareIte]].killTime ~= nil) then
            SendAddonMessage(syncChanNamePrefix, "[".. syncChanNamePrefix .. "]" .. GetRareIdString(mobIDs[syncRareIte]) .. "_death_" .. tostring(mobs[mobIDs[syncRareIte]].killTime), "WHISPER", sender)
          end
        end
      -- Sync response
      else
        ProcessSyncChatMsg(message)
        -- RefreshText(self, 100)
      end
    end
  elseif (strtrim(prefix) == syncRSChanNamePrefix) then
    if (channel == "WHISPER") then
      ProcessRSSyncChatMsg(message)
    end
  end
end

-- Channel roster update handler (used for requesting syncs)
local function ProcessChanRosterUpdate(chanID)
  local name, _, _, _, _, _, _, _, _ = GetChannelDisplayInfo(chanID)
  
  --print ("Processing roster update for ", name)
  if (name == syncChanNamePrefix) then
    -- Check if we need to sync
    if (needSync) then
      mainFrame:SetScript("OnUpdate", nil)
      needSync = false
      
      local chanOwner = GetChannelOwner(chanID)
      if (chanOwner ~= nil) then
        SendAddonMessage(syncChanNamePrefix, "Sync", "WHISPER", chanOwner)
      end
    end
  elseif (name == syncRSChanNamePrefix) then
    -- Check if we need to sync with RC
    if (needRCSync) then
      mainFrame:SetScript("OnUpdate", nil)
      needRCSync = false
      
      local chanOwner = GetChannelOwner(chanID)
      if (chanOwner ~= nil) then
        SendAddonMessage(syncRSChanNamePrefix, "GetStatus", "WHISPER", chanOwner)
      end
    end
  end
  
  if ((needSync   == false) or
      (needRCSync == false)) then
    mainFrame:SetScript("OnUpdate", nil)
  else
    mainFrame:SetScript("OnUpdate", JoinSyncChannels)
  end
end

-- Chat message handler for /1 and the sync channels
local function ProcessChatMsg(message, sender, language, channelString, target, flags, unknown, channelID, channelName, unknown, counter, guid)
  -- Handle request messages
  --if (channelID == 1) or (channelID == 4) then
  if (channelID == 1) then
    -- Request string
    if (strlower(string.sub(message,1,9)) == "raretime ") or
       (strlower(string.sub(message,1,10)) == "raretimer ") then
      -- Expected format =  ratime <npc name/id>
      ProcessRequest(message)
    end
  -- Addon Channel message
  elseif (channelName == syncChanNamePrefix) then
    if string.find(message, "["..syncChanNamePrefix.."]") then
      ProcessSyncChatMsg(message)
    end
  elseif (channelName == syncRSChanNamePrefix) then
    if string.find(message, "[" .. syncRSChanNamePrefix .. "]") then
      ProcessRSSyncChatMsg(message)
    end
  end
end

-- Zone changing event handler
local function EnableOrDisable()
  -- MapAreaID 951 = Timeless Isles
  if (GetCurrentMapAreaID() == 951) then
    mainFrame:RegisterEvent("CHAT_MSG_CHANNEL")
    mainFrame:RegisterEvent("CHAT_MSG_ADDON")
    mainFrame:RegisterEvent("PLAYER_TARGET_CHANGED")
    mainFrame:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    mainFrame:RegisterEvent("PLAYER_REGEN_DISABLED")
    mainFrame:RegisterEvent("PLAYER_REGEN_ENABLED")
    
    mainFrame:SetScript("OnUpdate", JoinSyncChannels)
    
    rareFrame:Show()
  else
    mainFrame:UnregisterEvent("CHAT_MSG_CHANNEL")
    mainFrame:UnregisterEvent("CHAT_MSG_ADDON")
    mainFrame:UnregisterEvent("PLAYER_TARGET_CHANGED")
    mainFrame:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
    mainFrame:UnregisterEvent("PLAYER_REGEN_DISABLED")
    mainFrame:UnregisterEvent("PLAYER_REGEN_ENABLED")
    
    mainFrame:SetScript("OnUpdate", nil)
    
    rareFrame:Hide()
    needSync = true
    needRCSync = true
  end
end

-- Player into combat event handler
local function HideFrameInCombat()
  if (RAREAN.Options.HideInCombat == true) then
    rareFrame:Hide()
  end
end

-- Player out of combat event handler
local function ShowFrameAfterCombat()
  if (RAREAN.Options.HideInCombat == true) then
    rareFrame:Show()
  end
end

-- Loads the saved variables when login and sets the defauls as needed
local function LoadSavedVariables()
  if (RAREAN.Options == nil) then
    RAREAN.Options = CopyTable(OptionsDefault)
  end
  
  if (RAREAN.Options.RespondToRequests == nil) then
    RAREAN.Options.RespondToRequests = OptionsDefault.RespondToRequests
  end

  if (RAREAN.Options.ReportFindsToGeneral == nil) then
    RAREAN.Options.ReportFindsToGeneral = OptionsDefault.ReportFindsToGeneral
  end

  if (RAREAN.Options.HideInCombat == nil) then
    RAREAN.Options.HideInCombat = OptionsDefault.HideInCombat
  end

  if (RAREAN.Options.ShowAllOnMouseOver == nil) then
    RAREAN.Options.ShowAllOnMouseOver = OptionsDefault.ShowAllOnMouseOver
  end

  if (RAREAN.Options.SetTomTomWaypoints == nil) then
    RAREAN.Options.SetTomTomWaypoints = OptionsDefault.SetTomTomWaypoints
  end

  if (RAREAN.Options.RareOrderType == nil) then
    RAREAN.Options.RareOrderType = OptionsDefault.RareOrderType
  end

  if (RAREAN.Options.TrackingType == nil) then
    RAREAN.Options.TrackingType = OptionsDefault.TrackingType
  end
end

-- Main event handler
local function MainEvents(mainFrame, event, ...)
  if (event == "COMBAT_LOG_EVENT_UNFILTERED") then
    local _,event_type,_,_,_,_,_,guid,name = select(1, ...)

    if event_type == "UNIT_DIED" then
      ProcessDeathLog(name, guid)
    end
  elseif (event == "PLAYER_TARGET_CHANGED") then
    local guid = UnitGUID("target")
    if guid and mobs[tonumber(guid:sub(6, 10), 16)] and not UnitIsDead("target") then
      local id = tonumber(guid:sub(6, 10), 16)
      local name = strtrim(UnitName("target"))
      mobs[id].killTime = nil -- Reset the kill time to Alive
      mobs[id].npcName = name
      SendTargetMessage(id)
    end
  elseif (event == "CHAT_MSG_CHANNEL") then
    ProcessChatMsg(...)
  elseif (event == "CHAT_MSG_ADDON") then
    ProcessAddonMsg(...)
  elseif (event == "PLAYER_REGEN_DISABLED") then
    HideFrameInCombat()
  elseif (event == "PLAYER_REGEN_ENABLED") then
    ShowFrameAfterCombat()
  elseif (event == "PLAYER_ENTERING_WORLD" or event == "ZONE_CHANGED_NEW_AREA") then
    EnableOrDisable()
  elseif (event == "CHANNEL_ROSTER_UPDATE") then
    ProcessChanRosterUpdate(...)
  elseif (event == "ADDON_LOADED") then
    LoadSavedVariables()
  end
end

mainFrame:RegisterEvent("PLAYER_ENTERING_WORLD")
mainFrame:RegisterEvent("ZONE_CHANGED_NEW_AREA")
mainFrame:RegisterEvent("CHANNEL_ROSTER_UPDATE")
mainFrame:RegisterEvent("ADDON_LOADED")
mainFrame:SetScript("OnEvent", MainEvents)

-- === OPTION PANELS START === --

-- Generic CheckBox maker
local function CreateCheckButton(parent, checkBoxName, posX, posY, displayText)
	local checkbutton = CreateFrame("CheckButton", checkBoxName, parent, "UICheckButtonTemplate");
	checkbutton:SetPoint("TOPLEFT", posX, posY);
  checkbutton:SetWidth(25)
  checkbutton:SetHeight(25)
	getglobal(checkbutton:GetName() .. 'Text'):SetText(displayText);

	return checkbutton;
end

  -- Set the options when showing the interface options menu
  function OptionsFrameSettingsLoad()
    mainFrame.optionsFrame.chkRespondToRequest:SetChecked(RAREAN.Options.RespondToRequests)
    mainFrame.optionsFrame.chkReportFindsToGeneral:SetChecked(RAREAN.Options.ReportFindsToGeneral)
    mainFrame.optionsFrame.chkHideInCombat:SetChecked(RAREAN.Options.HideInCombat)
    --mainFrame.optionsFrame.chkSetTomTom:SetChecked(RAREAN.Options.SetTomTomWaypoints)
  end

-- UI Options panels
  -- Main Panel
  mainFrame.optionsFrame = CreateFrame("Frame", "mainFrame.optionFrame", UIParent );
  mainFrame.optionsFrame.name = "RareAnnouncer"
  mainFrame.optionsFrame.refresh = OptionsFrameSettingsLoad

    -- Respond to Request option
    mainFrame.optionsFrame.chkRespondToRequest = CreateCheckButton(mainFrame.optionsFrame, "chkRespondToRequest", 20, -20, "Respond to requests")
    mainFrame.optionsFrame.chkRespondToRequest.tooltip = "Respond to raretime/raretimer requests"
    mainFrame.optionsFrame.chkRespondToRequest:SetScript("OnClick", 
      function()
       if (mainFrame.optionsFrame.chkRespondToRequest:GetChecked() == 1) then
         RAREAN.Options.RespondToRequests = true
       else
         RAREAN.Options.RespondToRequests = false
       end
      end
    );
    -- Report finds/death to General
    mainFrame.optionsFrame.chkReportFindsToGeneral = CreateCheckButton(mainFrame.optionsFrame, "chkReportFindsToGeneral", 20, -42, "Report rare finds and deaths to general chat")
    mainFrame.optionsFrame.chkReportFindsToGeneral.tooltip = "Announce when you target a rare or when a rare dies"
    mainFrame.optionsFrame.chkReportFindsToGeneral:SetScript("OnClick", 
      function()
       if (mainFrame.optionsFrame.chkReportFindsToGeneral:GetChecked() == 1) then
         RAREAN.Options.ReportFindsToGeneral = true
       else
         RAREAN.Options.ReportFindsToGeneral = false
       end
      end
    );
    -- Hide in combat
    mainFrame.optionsFrame.chkHideInCombat = CreateCheckButton(mainFrame.optionsFrame, "chkHideInCombat", 20, -64, "Hide tracking window while in combat")
    mainFrame.optionsFrame.chkHideInCombat.tooltip = "Announce when you target a rare or when a rare dies"
    mainFrame.optionsFrame.chkHideInCombat:SetScript("OnClick", 
      function()
       if (mainFrame.optionsFrame.chkHideInCombat:GetChecked() == 1) then
         RAREAN.Options.HideInCombat = true
       else
         RAREAN.Options.HideInCombat = false
       end
      end
    );
    -- Set TomTom Waypoints
    -- mainFrame.optionsFrame.chkSetTomTom = CreateCheckButton(mainFrame.optionsFrame, "chkSetTomTom", 20, -96, "Set TomTom waypoint when a rare is found that you are not targeting")
    -- mainFrame.optionsFrame.chkSetTomTom.tooltip = "Set TomTom waypoint when a rare is found that you are not targeting"
    -- mainFrame.optionsFrame.chkSetTomTom:SetScript("OnClick", 
    --   function()
    --    if (mainFrame.optionsFrame.chkSetTomTom:GetChecked() == 1) then
    --      RAREAN.Options.SetTomTomWaypoints = true
    --    else
    --      RAREAN.Options.SetTomTomWaypoints = false
    --    end
    --   end
    -- );

    -- mainFrame.optionsFrame.rbnTrackOne = CreateFrame("CheckButton", "rbnTrackOne", mainFrame.optionsFrame, "UIRadioButtonTemplate");
    -- mainFrame.optionsFrame.rbnTrackOne:SetPoint("TOPLEFT", 22, -112);
    -- mainFrame.optionsFrame.rbnTrackOne:SetHeight(20)
    -- mainFrame.optionsFrame.rbnTrackOne:SetWidth(20)
	  -- getglobal(mainFrame.optionsFrame.rbnTrackOne:GetName() .. 'Text'):SetText("First tracking option");
    
    -- RAREAN.Options
    -- DONE RespondToRequests,
    -- DONE ReportFindsToGeneral,
    -- DONE HideInCombat,
    -- ShowAllOnMouseOver,
    -- DONE SetTomTomWaypoints,
    -- RareOrderType,
    -- TrackingType,
    -- TrackingRares = {}
  InterfaceOptions_AddCategory(mainFrame.optionsFrame);

-- === OPTION PANELS START === --

mainFrame.updateFrame = CreateFrame("Frame", "mainFrame.updateFrame", mainFrame)
mainFrame.updateFrame:SetScript("OnUpdate", OverallUpdater)
OverallUpdater(mainFrame, 10)


















